#define _SVID_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "cuda_utils.h"

void MatrixProd(float* a,float* b,float* c,int m,int n,int p) {
  for(int i=0;i<m;i++) {
    size_t a_line_index=i*n;
    for(int j=0;j<p;j++) {
      float sum=0;
      for(int k=0;k<n;k++) {
        sum+=a[a_line_index+k]*b[k*p+j];
      }
      c[i*p+j]=sum;
    }
  }
}

// CPU code //

int main() {

  srand48(time(NULL));

  // define sizes //
  const int size_m = 200;
  const int size_n = 300;
  const int size_p = 400;

  const size_t matrix_a_size = size_m*size_n*sizeof(float);
  const size_t matrix_b_size = size_n*size_p*sizeof(float);
  const size_t matrix_c_size = size_m*size_p*sizeof(float);

  // allocate memory on cpu //
  float* matrix_a = malloc(matrix_a_size);
  float* matrix_b = malloc(matrix_b_size);
  float* matrix_c = malloc(matrix_c_size);
  
  // initialize matrixs //
  for(int i=0;i<size_m*size_n;i++) {
    matrix_a[i]=drand48();
  }
  for(int i=0;i<size_n*size_p;i++) {
    matrix_b[i]=drand48();
  }
  for(int i=0;i<size_m*size_p;i++) {
    matrix_c[i]=0;
  }

  Timer timer;
  initTimer(&timer);

  // computation function call //
  MatrixProd(matrix_a,matrix_b,matrix_c,size_m,size_n,size_p);

  double duration = getTimer(&timer);

  for(int i=0;i<size_m;i++) {
    for(int j=0;j<size_p;j++) {
      fprintf(stderr,"%e ",matrix_c[i*size_p+j]);
    }
    fprintf(stderr,"\n");
  }

  fprintf(stderr,"duration = %e s, performance = %e FLOPS\n",duration,
          2.*size_m*size_n*size_p/duration);

  // free memory //
  free(matrix_a);
  free(matrix_b);
  free(matrix_c);

  // exit //
  return 0;
}
