module device_code

use cudafor

integer, parameter :: grid_dim = 90

contains

! CUDA Kernel !

attributes(global) subroutine ScalarProduct(a,b,result,size)
  implicit none
  integer, dimension(0:size-1) :: a,b
  integer, intent(out) :: result
  integer, value :: size
  integer :: threadSum
  integer, shared :: blockSum
  integer :: i
  if(threadIdx%x==1) then
    blockSum=0
  end if
  call syncthreads()
  threadSum=0
  do i=(threadIdx%x-1)+(blockIdx%x-1)*blockDim%x,size-1,blockDim%x*gridDim%x
    threadSum=threadSum+a(i)*b(i)
  end do
  ! TODO: add both reduction stages here !
end subroutine ScalarProduct

end module device_code

program exercise14

use cudafor
use device_code

implicit none

! define vector size !
integer, parameter :: vector_size = 10000
integer, dimension(0:vector_size-1) :: vector_a_cpu
integer, dimension(0:vector_size-1) :: vector_b_cpu
integer :: result_cpu,sum
real :: random
integer, dimension(0:vector_size-1), device :: vector_a_gpu
integer, dimension(0:vector_size-1), device :: vector_b_gpu
integer, device :: result_gpu
real(kind=8) :: timer1,timer2,duration
integer :: i
logical :: success
integer :: error

! CPU code !

call random_seed()

! initialize cpu input vector to 0 !
do i=0,vector_size-1
  call random_number(random)
  vector_a_cpu(i)=random*100
  call random_number(random)
  vector_b_cpu(i)=random*100
end do

! copy input vectors a,b to gpu !
vector_a_gpu=vector_a_cpu
vector_b_gpu=vector_b_cpu
! initialize gpu result to 0 !
result_gpu=0

call cpu_time(timer1)
! cuda kernel call !
call ScalarProduct<<<grid_dim,32>>>(vector_a_gpu,vector_b_gpu,result_gpu,vector_size)
error = cudaThreadSynchronize()
call cpu_time(timer2)
duration=timer2-timer1

! copy output vector from gpu to cpu !
result_cpu=result_gpu

! verify result !
sum=0
do i=0,vector_size-1
  sum=sum+vector_a_cpu(i)*vector_b_cpu(i)
end do
if (abs(sum-result_cpu)>1e-5) then
  print *,'computation result is wrong: expected result=',sum, &      
        'cuda result=', result_cpu
else
  print *,'computation result is correct.'
end if

print *,'duration=',duration,' seconds, performance = ',vector_size*2./duration,' FLOPS'
print *,'memory bandwidth = ',vector_size*8./duration,' bytes/s'

end program exercise14
