module device_code

contains

! CUDA Kernel !

attributes(global) subroutine VectorSum(a,b,c,size)
  implicit none
  real, dimension(0:size-1) :: a,b,c
  integer, value :: size
  integer :: i
  do i=(threadIdx%x-1)+(blockIdx%x-1)*blockDim%x,size-1,blockDim%x*gridDim%x
    c(i)=a(i)+b(i)
  end do
end subroutine VectorSum

end module device_code

program exercise04

use cudafor
use device_code

implicit none

! define vector size !
integer, parameter :: vector_size = 10000000
integer, parameter ::  offset = 8 ! offset for nonaligned access !
real, dimension(0:vector_size-1) :: vector_a_cpu
real, dimension(0:vector_size-1) :: vector_b_cpu
real, dimension(0:vector_size-1) :: vector_c_cpu
real, dimension(0:vector_size-1+offset), device :: vector_a_gpu
real, dimension(0:vector_size-1+offset), device :: vector_b_gpu
real, dimension(0:vector_size-1), device :: vector_c_gpu
real(kind=8) :: timer1,timer2,duration
integer :: i
logical :: success
integer :: error

! CPU code !

call random_seed()

! initialize cpu input vector to 0 !
do i=0,vector_size-1
  call random_number(vector_a_cpu(i))
  call random_number(vector_b_cpu(i))
end do
! initialize cpu output vector to 0 !
do i=0,vector_size-1
  vector_c_cpu(i)=0
end do

! copy input vectors a,b to gpu !
vector_a_gpu(offset:vector_size+offset-1)=vector_a_cpu
vector_b_gpu(offset:vector_size+offset-1)=vector_b_cpu
! initialize gpu output vector to 0 !
vector_c_gpu=vector_c_cpu

call cpu_time(timer1)
! cuda kernel call !
call VectorSum<<<128,64>>>(vector_a_gpu(offset:vector_size+offset-1),vector_b_gpu(offset:vector_size+offset-1),vector_c_gpu,vector_size)
error = cudaThreadSynchronize()
call cpu_time(timer2)
duration=timer2-timer1

! copy output vector from gpu to cpu !
vector_c_cpu=vector_c_gpu

! verify results !
success=.true.
do i=0,vector_size-1
  if (abs(vector_a_cpu(i)+vector_b_cpu(i)-vector_c_cpu(i))>1e-5) then
    print *,'computation result is wrong: index=',i, &
        ' expected result=',vector_a_cpu(i)*vector_b_cpu(i), &      
        'cuda result=', vector_c_cpu(i)
    success=.false.
    exit
  end if
end do
if (success) then
  print *,'computation result is correct.'
end if

print *,'duration=',duration,' seconds, performance = ',vector_size/duration,' FLOPS'
print *,'memory bandwidth = ',vector_size*12./duration,' bytes/s'

end program exercise04
