module device_code

contains

! CUDA Kernel !

attributes(global) subroutine ScalarProduct(a,b,result,size)
  implicit none
  real, dimension(0:size-1) :: a,b
  real :: result
  integer, value :: size
  real :: sum
  real, shared, dimension(0:31) :: sums
  integer :: i
  sum=0
  do i=(threadIdx%x-1),size-1,blockDim%x
     sum=sum+a(i)*b(i)
  end do
  if (threadIdx%x<=16) then
    sums(threadIdx%x-1)=sum
  end if
  call syncthreads()
  if (threadIdx%x>16) then
    sums(threadIdx%x-1-16)=sums(threadIdx%x-1-16)+sum
  end if
  call syncthreads()
  if (threadIdx%x<=8) then
    sums(threadIdx%x-1)=sums(threadIdx%x-1)+sums(threadIdx%x-1+8)
  end if
  call syncthreads()
  if (threadIdx%x<=4) then
    sums(threadIdx%x-1)=sums(threadIdx%x-1)+sums(threadIdx%x-1+4)
  end if
  call syncthreads()
  if (threadIdx%x==1) then
    result=(sums(0)+sums(1))+(sums(2)+sums(3))
  end if
end subroutine ScalarProduct

end module device_code

program exercise05

use cudafor
use device_code

implicit none

! define vector size !
integer, parameter :: vector_size = 10000
real, dimension(0:vector_size-1) :: vector_a_cpu
real, dimension(0:vector_size-1) :: vector_b_cpu
real :: result_cpu,sum
real, dimension(0:vector_size-1), device :: vector_a_gpu
real, dimension(0:vector_size-1), device :: vector_b_gpu
real, device :: result_gpu
real(kind=8) :: timer1,timer2,duration
integer :: i
logical :: success
integer :: error

! CPU code !

call random_seed()

! initialize cpu input vector to 0 !
do i=0,vector_size-1
  vector_a_cpu(i)=1./(1.+i)
  vector_b_cpu(i)=1.+i
end do

! copy input vectors a,b to gpu !
vector_a_gpu=vector_a_cpu
vector_b_gpu=vector_b_cpu
! initialize gpu result to 0 !
result_gpu=0

call cpu_time(timer1)
! cuda kernel call !
call ScalarProduct<<<1,32>>>(vector_a_gpu,vector_b_gpu,result_gpu,vector_size)
error = cudaThreadSynchronize()
call cpu_time(timer2)
duration=timer2-timer1

! copy output vector from gpu to cpu !
result_cpu=result_gpu

! verify result !
sum=0;
do i=0,vector_size-1
  sum=sum+vector_a_cpu(i)*vector_b_cpu(i)
end do
if (abs(sum-result_cpu)>1e-5) then
  print *,'computation result is wrong: expected result=',sum, &      
      'cuda result=', result_cpu
else
  print *,'computation result is correct.'
end if

print *,'duration=',duration,' seconds, performance = ',vector_size*2./duration,' FLOPS'
print *,'memory bandwidth = ',vector_size*8./duration,' bytes/s'

end program exercise05
