module device_code

integer, parameter :: grid_dim = 90

contains

! CUDA Kernel !

attributes(global) subroutine ScalarProduct1(a,b,results,size)
  implicit none
  real, dimension(0:size-1) :: a,b
  real, dimension(0:gridDim%x-1) :: results
  integer, value :: size
  real, shared, dimension(0:31) :: sums
  real :: sum
  integer :: i
  sum=0
  do i=(threadIdx%x-1)+(blockIdx%x-1)*blockDim%x,size-1,blockDim%x*gridDim%x
     sum=sum+a(i)*b(i)
  end do
  if (threadIdx%x<=16) then
    sums(threadIdx%x-1)=sum
  end if
  call syncthreads()
  if (threadIdx%x>16) then
    sums(threadIdx%x-1-16)=sums(threadIdx%x-1-16)+sum
  end if
  call syncthreads()
  if (threadIdx%x<=8) then
    sums(threadIdx%x-1)=sums(threadIdx%x-1)+sums(threadIdx%x-1+8)
  end if
  call syncthreads()
  if (threadIdx%x<=4) then
    sums(threadIdx%x-1)=sums(threadIdx%x-1)+sums(threadIdx%x-1+4)
  end if
  call syncthreads()
  if (threadIdx%x==1) then
    results(blockIdx%x-1)=(sums(0)+sums(1))+(sums(2)+sums(3))
  end if
end subroutine ScalarProduct1

attributes(global) subroutine ScalarProduct2(results)
  real, dimension(0:grid_dim-1) :: results
  real :: sum
  integer :: i
  sum=0
  do i = 0,grid_dim-1
    sum=sum+results(i)
  end do
  results(0)=sum;
end subroutine ScalarProduct2

end module device_code

program exercise06

use cudafor
use device_code

implicit none

! define vector size !
integer, parameter :: vector_size = 10000
real, dimension(0:vector_size-1) :: vector_a_cpu
real, dimension(0:vector_size-1) :: vector_b_cpu
real :: result_cpu,sum
real, dimension(0:vector_size-1), device :: vector_a_gpu
real, dimension(0:vector_size-1), device :: vector_b_gpu
real, device, dimension(0:grid_dim-1) :: results_gpu
real(kind=8) :: timer1,timer2,duration
integer :: i
logical :: success
integer :: error

! CPU code !

call random_seed()

! initialize cpu input vector to 0 !
do i=0,vector_size-1
  vector_a_cpu(i)=1./(1.+i)
  vector_b_cpu(i)=1.+i
end do

! copy input vectors a,b to gpu !
vector_a_gpu=vector_a_cpu
vector_b_gpu=vector_b_cpu
! initialize gpu result to 0 !
results_gpu=0

call cpu_time(timer1)
! cuda kernel call !
call ScalarProduct1<<<grid_dim,32>>>(vector_a_gpu,vector_b_gpu,results_gpu,vector_size)
call ScalarProduct2<<<1,1>>>(results_gpu)
error = cudaThreadSynchronize()
call cpu_time(timer2)
duration=timer2-timer1

! copy output vector from gpu to cpu !
result_cpu=results_gpu(0)

! verify result !
sum=0;
do i=0,vector_size-1
  sum=sum+vector_a_cpu(i)*vector_b_cpu(i)
end do
if (abs(sum-result_cpu)>1e-5) then
  print *,'computation result is wrong: expected result=',sum, &      
      'cuda result=', result_cpu
else
  print *,'computation result is correct.'
end if

print *,'duration=',duration,' seconds, performance = ',vector_size*2./duration,' FLOPS'
print *,'memory bandwidth = ',vector_size*8./duration,' bytes/s'

end program exercise06
